import { ObjectId } from "mongodb";
import MissionType from "../../../types/MissionType";
import DBFleet from "../../../types/db/DBFleet";
import { updateFleet } from "../../db/fleet";
import { Planet } from "./PlanetManager";
import SystemManager, { System } from "./SystemManager";
import { sendMail } from "../../db/mails";
import { Sector } from "./LocationManager";
import { getRandomInRange, weightedRandom } from "../../utils/math";
import { getAllShips } from "../../db/ships";

export type Fleet = {
    id: ObjectId,
    source: Planet | SystemManager,
    destination: Planet | SystemManager | Sector,
    departureTime: Date,
    arrivalTime: Date,
    returning: boolean,
    mission: MissionType,
    ships: Array<{ id: string, amount: number }>,
    cargo: Array<{ id: string, amount: number }>
}

export default class FleetManager {
    data: Fleet;

    constructor(id: ObjectId, source: Planet | SystemManager, destination: Planet | SystemManager | Sector, departureTime: Date, arrivalTime: Date, returning: boolean, mission: MissionType, ships: Array<{ id: string, amount: number }>, cargo: Array<{ id: string, amount: number }>) {
        this.data = {
            id,
            source,
            destination,
            departureTime,
            arrivalTime,
            returning,
            mission,
            ships,
            cargo
        }
    }

    isSourcePlanet(): this is { data: { source: Planet } } {
        return 'resources' in this.data.source;
    }

    isSourceSystem(): this is { data: { source: SystemManager } } {
        return 'structures' in this.data.source;
    }

    isDestinationPlanet(): this is { data: { destination: Planet } } {
        return 'resources' in this.data.destination;
    }

    isDestinationSystem(): this is { data: { destination: SystemManager } } {
        return 'structures' in this.data.destination;
    }

    getSourceOwner() {
        if(this.isSourcePlanet()) return this.data.source.system.data.ownedBy;
        if(this.isSourceSystem()) return this.data.source.data.ownedBy;
    }

    getDestinationOwner() {
        if(this.isDestinationPlanet()) return this.data.destination.system.data.ownedBy;
        if(this.isDestinationSystem()) return this.data.destination.data.ownedBy;
    }

    async checkStatus(): Promise<{ finished: boolean, fleet: Fleet }> {
        if(this.data.arrivalTime.getTime() < Date.now()) {
            const finished = await this.finish();
            return { finished, fleet: this.data };
        }

        return { finished: false, fleet: this.data };
    }

    async finish() {
        if(this.data.returning) {
            for(const ship of this.data.ships) {
                this.data.source.ships.addShips(ship.id, ship.amount);
            }
            await this.data.source.resources.updateAmount(this.data.cargo);
            await this.data.source.ships.sync();
            await this.data.source.resources.sync();
            await sendMail(
                null,
                this.data.source instanceof SystemManager ? this.data.source.data.ownedBy.id : this.data.source.system.data.ownedBy.id,
                this.data.arrivalTime,
                "Fleet Returned",
                `Your fleet from ${this.data.destination instanceof SystemManager ? `${this.data.destination.data.name} system` : `${this.data.destination.name} ${"expedition" in this.data.destination ? "sector" : "planet"}`} has returned.\n
                Ships: ${this.data.ships.map(ship => `${ship.amount} ${ship.id}`).join(', ')}\n
                Cargo: ${this.data.cargo.length > 0 ? this.data.cargo.map(cargo => `${cargo.amount} ${cargo.id}`).join(', ') : 'None'}`
            );
            return true;
        } else {
            switch(this.data.mission) {
                case 'ATTACK':
                    return false;
                case 'TRANSPORT':
                    await (this.data.destination as Planet | SystemManager).resources.updateAmount(this.data.cargo);
                    await (this.data.destination as Planet | SystemManager).resources.sync();
                    const cargo = JSON.parse(JSON.stringify(this.data.cargo)) as Array<{ id: string, amount: number }>;
                    this.data.cargo = [];
                    const arrived = new Date(this.data.arrivalTime);
                    await this.initiateReturn();
                    await sendMail(
                        null,
                        this.data.source instanceof SystemManager ? this.data.source.data.ownedBy.id : this.data.source.system.data.ownedBy.id,
                        arrived,
                        "Fleet Arrived",
                        `Your fleet has arrived at ${this.data.destination instanceof SystemManager ? `${this.data.destination.data.name} system` : `planet ${this.data.destination.name}`}.\n
                        Ships: ${this.data.ships.map(ship => `${ship.amount} ${ship.id}`).join(', ')}\n
                        Cargo delivered: ${cargo.length > 0 ? cargo.map(cargo => `${cargo.amount} ${cargo.id}`).join(', ') : 'None'}\n
                        Fleet will return at ${this.data.arrivalTime}`
                    );
                    return false;
                case 'TRANSFER':
                    await (this.data.destination as Planet | SystemManager).resources.updateAmount(this.data.cargo);
                    await (this.data.destination  as Planet | SystemManager).resources.sync();
                    for(const ship of this.data.ships) {
                        (this.data.destination as Planet | SystemManager).ships.addShips(ship.id, ship.amount);
                    }
                    await (this.data.destination as Planet | SystemManager).ships.sync();
                    await sendMail(
                        null,
                        this.data.source instanceof SystemManager ? this.data.source.data.ownedBy.id : this.data.source.system.data.ownedBy.id,
                        this.data.arrivalTime,
                        "Fleet Arrived",
                        `Your fleet has arrived at ${this.data.destination instanceof SystemManager ? `${this.data.destination.data.name} system` : `planet ${this.data.destination.name}`}.\n
                        Ships: ${this.data.ships.map(ship => `${ship.amount} ${ship.id}`).join(', ')}\n
                        Cargo delivered: ${this.data.cargo.length > 0 ? this.data.cargo.map(cargo => `${cargo.amount} ${cargo.id}`).join(', ') : 'None'}\n
                        Ships will stay at the destination.`
                    );
                    return true;
                case 'EXPEDITION':
                    await this.expeditionResults();

                    return false;
            }
        }
    }

    async initiateReturn() {
        this.data.returning = true;
        this.data.departureTime = new Date(this.data.arrivalTime);
        this.data.arrivalTime = new Date(this.data.departureTime.getTime() + 1000 * 30);
        await this.sync();
    }

    private async sendMail(description: string) {
        await sendMail(
            null,
            this.data.source instanceof SystemManager ? this.data.source.data.ownedBy.id : this.data.source.system.data.ownedBy.id,
            this.data.arrivalTime,
            "Expedition Results",
            description
        );
    }

    async expeditionResults() {
        const expeditionRandom = Math.random(); //TODO: make use of "expedition" from DBSector

        const allShips = await getAllShips();
        const shipCount = this.data.ships.reduce((acc, ship) => acc += ship.amount, 0);
        const totalShipCapacity = this.data.ships.reduce((acc, ship) => acc + (allShips.find(s => s.id === ship.id)?.capacity.solid ?? 0) * ship.amount, 0);
        const currentCargo = this.data.cargo.reduce((acc, res) => acc + res.amount, 0);

        const expeditionShipPositiveRandom = weightedRandom(shipCount > 100_000 ? 100_000 : shipCount);
        const valueAdded = Math.floor(Math.pow(Math.E * expeditionShipPositiveRandom, 4 * Math.PI) + 10)

        if(expeditionRandom < 0.02) { // 2% chance; lost all ships, black hole
            this.data.ships = [];
            await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector encountered a black hole. All ships were lost.`);
            return;
        }

        if(expeditionRandom < 0.1) { // 8% chance; found ships
            const ships: { id: string, amount: number }[] = [];
            ships.push({ id: 'transporter', amount: getRandomInRange(0, 20) + Math.floor(valueAdded ) });
            ships.push({ id: 'fighter', amount: getRandomInRange(0, 20) + Math.floor(valueAdded ) });

            for(const s of ships) {
                const ship = this.data.ships.find((sh => sh.id === s.id));
                if(!ship) this.data.ships.push(s);
                else ship.amount += s.amount;
            }

            await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector encountered abandoned shipyard. Following ships were added to the fleet:\n${ships.map(s => `${s.id} - ${s.amount}`).join(', ')}`);
            await this.initiateReturn();
            return;
        }

        if(expeditionRandom < 0.2) { // 10% chance; found resources           
            let totalCapacityUsed = 0;
            let resources = [];

            const resToAdd = valueAdded * 3 + currentCargo > totalShipCapacity ? totalShipCapacity / 3 : valueAdded;

            do {
                resources = [{
                    id: "coal",
                    amount: getRandomInRange(0, 10000) + resToAdd
                }, {
                    id: "iron",
                    amount: getRandomInRange(0, 10000) + resToAdd
                }, {
                    id: "gold",
                    amount: getRandomInRange(0, 10000) + resToAdd
                }];

                totalCapacityUsed = resources.reduce((acc, res) => acc + res.amount, 0);
            } while((totalCapacityUsed + currentCargo) > totalShipCapacity);

            for(const res of resources) {
                const resource = this.data.cargo.find(r => r.id === res.id);
                if(!resource) this.data.cargo.push(res);
                else resource.amount += res.amount;
            }

            await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector encountered resource-rich asteroid. Following resources were added to the cargo inventory:\n${resources.map(r => `${r.id} - ${r.amount}`).join('\n')}`);
            await this.initiateReturn();
            return;
        }

        if(expeditionRandom < 0.35) { // 15% chance; pirates/aliens
            //TODO: implement fight mechanic

            await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector encountered drunk pirates. After attempting to communicate with them, they attacked your fleet.`)
            await this.initiateReturn();
            return;
        }

        if(expeditionRandom < 0.36) { // 1% chance; rich rouge planet
            const oneResourceMax = totalShipCapacity / 3;

            let currentCargo = this.data.cargo.reduce((acc, res) => {
                return acc + res.amount;
            }, 0);

            const addedResources: { id: string, amount: number }[] = [{
                id: "coal",
                amount: 0
            }, {
                id: "iron",
                amount: 0
            }, {
                id: "gold",
                amount: 0
            }];

            while(currentCargo < totalShipCapacity) {
                if((addedResources.find(r => r.id === "coal")?.amount ?? 0) < oneResourceMax) {
                    (addedResources.find(r => r.id === "coal") as { amount: number }).amount++;
                    currentCargo++;
                    if(currentCargo >= totalShipCapacity) break;
                }

                if((addedResources.find(r => r.id === "iron")?.amount ?? 0) < oneResourceMax) {
                    (addedResources.find(r => r.id === "iron") as { amount: number }).amount++;
                    currentCargo++;
                    if(currentCargo >= totalShipCapacity) break;
                }

                if((addedResources.find(r => r.id === "gold")?.amount ?? 0) < oneResourceMax) {
                    (addedResources.find(r => r.id === "gold") as { amount: number }).amount++;
                    currentCargo++;
                    if(currentCargo >= totalShipCapacity) break;
                }
            }

            for(const res of addedResources) {
                const resource = this.data.cargo.find(r => r.id === res.id);
                if(!resource) this.data.cargo.push(res);
                else resource.amount += res.amount;
            }

            await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector encountered resource-rich rouge planet. Your fleet could not extract all resources. Following resources were added to the cargo inventory:\n${addedResources.map(r => `${r.id} - ${r.amount}`).join('\n')}`);
            await this.initiateReturn();
            return;
        } 

        await this.sendMail(`Your expedition to ${(this.data.destination as Sector).name} sector scanned the sector for a long time, yet it haven't found anything.`);
        await this.initiateReturn();
        return;
    }

    async sync() {
        const source = this.data.source instanceof SystemManager ? this.data.source.data._id : this.data.source._id;
        const destination = this.data.destination instanceof SystemManager ? this.data.destination.data._id : this.data.destination._id;

        const data: DBFleet = {
            _id: this.data.id,
            source,
            destination,
            departureTime: this.data.departureTime,
            arrivalTime: this.data.arrivalTime,
            returning: this.data.returning,
            mission: this.data.mission,
            ships: this.data.ships,
            cargo: this.data.cargo
        }

        await updateFleet(data);
    }
}